# Script to make GET or POST requests to psd2.holvi.com with HTTP signature and all required headers.
# Python 2.7 and Python 3.5+ compatible.
# requires requests and cryptography packages.
import time
import base64
import datetime
import hashlib
import email.utils

from collections import OrderedDict

import requests

from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import load_pem_private_key

try:
	from urllib.parse import urlparse
except ImportError:
	from urlparse import urlparse


def get_common_headers(prepared_req):
	host = urlparse(prepared_req.url).netloc
	date = email.utils.formatdate(time.time(), usegmt=True)
	path = requests.models.RequestEncodingMixin.path_url.fget(prepared_req)
	request_target = '{} {}'.format(
		prepared_req.method.lower(),
		path,
	)
	headers = OrderedDict(
		(
			('(request-target)', request_target),
			('Host', host),
			('Date', date),
		)
	)
	return headers


def calculate_request_digest(prepared_req):
	digest = hashlib.sha256(
		prepared_req.body,
	).digest()
	return 'SHA-256=' + base64.b64encode(digest).decode()


def get_encrypted_signature(headers, private_key):
	string_to_sign = '\n'.join(
		'{}: {}'.format(name.lower(), value)
		for name, value in headers.items()
	)
	hasher = SHA256()
	raw_signature = private_key.sign(
		string_to_sign.encode(),
		padding=PKCS1v15(),
		algorithm=hasher,
	)
	return base64.b64encode(raw_signature).decode()


def get_signature_header(key_id, headers, private_key):
	encrypted_signature = get_encrypted_signature(headers, private_key)
	return (
		'keyId="{key_id}",'
		'algorithm="rsa-sha256",'
		'headers="{headers}",'
		'signature="{encrypted_sig}"'.format(
			key_id=key_id,
			headers=' '.join(name.lower() for name in headers),
			encrypted_sig=encrypted_signature,
		)
	)


def get_headers_with_content(prepared_req):
	headers = get_common_headers(prepared_req)
	content_type = prepared_req.headers.get('Content-Type', None)
	if content_type is None:
		raise ValueError(
			'Content-Type header is required for'
			'requests with content'
		)
	headers['Content-Type'] = content_type
	headers['Digest'] = calculate_request_digest(prepared_req)
	return headers


def get_headers(prepared_req, private_key,
				holvi_client_id, holvi_client_secret,
				):
	if prepared_req.method.lower() in ('post', 'patch', 'put'):
		headers = get_headers_with_content(prepared_req)
	else:
		headers = get_common_headers(prepared_req)
	headers['Signature'] = get_signature_header(
		holvi_client_id, headers, private_key,
	)
	headers['X-Holvi-Client-Id'] = holvi_client_id
	headers['X-Holvi-Client-Secret'] = holvi_client_secret
	return headers


def make_request(method, url,
				 holvi_client_id,
				 holvi_client_secret,
				 private_key, **kwargs
				 ):
	prepared_req = requests.Request(method, url, **kwargs).prepare()
	headers = get_headers(
		prepared_req,
		private_key,
		holvi_client_id,
		holvi_client_secret,
	)
	# (request-target) is used for generation, but
	# it's not a correct header to send
	headers.pop('(request-target)')
	prepared_req.headers.update(headers)
	session = requests.Session()
	return session.send(prepared_req)


def load_private_key():
	with open('eidas_cert_private_key.key', 'rb') as f:
		private_key = load_pem_private_key(
			f.read(),
			password=None,
			backend=default_backend(),
		)
	return private_key


# GET request

def make_get_request():
	private_key = load_private_key()
	response = make_request(
		method='GET',
		url='https://psd2.holvi.com/screen-scraping/v1/accounts/',
		holvi_client_id='testkeyid',
		holvi_client_secret='test_client_secret',
		private_key=private_key,
		params={'state': 'paid'},
	)
	assert response.status_code == 401


# POST request

def make_post_request():
	private_key = load_private_key()
	payload = {
		'iban': 'FI1979981000891140',
		'bic': 'HOLVFIHH',
		'receiver': {
			'name': 'BigBon',
		},
		'due_date': (
				datetime.datetime.utcnow() + datetime.timedelta(days=14)
		).date().isoformat(),
		'items': [
			{
				'description': 'Item description',
				'detailed_price': {
					'net': 100,
					'gross': 100,
				},
			},
		],
		'pool_handle': 'testing',
		'unstructured_reference': 'UNST-REF-666',
	}

	response = make_request(
		method='POST',
		url='https://psd2.holvi.com/screen-scraping/v1/testing/debt/',
		holvi_client_id='testkeyid',
		holvi_client_secret='test_client_secret',
		private_key=private_key,
		json=payload,
	)
	assert response.status_code == 401

if __name__ == '__main__':
	make_get_request()
	make_post_request()
